from sssekai.unity.AssetBundle import load_assetbundle
import os, json


def main_mvdata(args):
    from UnityPy.enums import ClassIDType

    cache_dir = args.cache_dir
    cache_dir = os.path.expanduser(cache_dir)
    cache_dir = os.path.abspath(cache_dir)
    os.chdir(cache_dir)
    mvdata_keys = os.listdir(cache_dir)
    mvdata_items = list()
    mvdata_lut = dict()
    for key in mvdata_keys:
        with open(key, "rb") as f:
            env = load_assetbundle(f)
            for obj in env.objects:
                if obj.type == ClassIDType.MonoBehaviour:
                    data = obj.read()
                    if data.name == "data":
                        index = len(mvdata_items)
                        mvdata_items.append(data.read_typetree())
                        mvdata_lut[str(data.name)] = index
                        mvdata_lut[str(data.id)] = index
                        break
    mvdata = mvdata_items[mvdata_lut[args.query]]
    print(json.dumps(mvdata, indent=4, ensure_ascii=False))



import os
from sssekai.unity.AssetBundle import load_assetbundle
from UnityPy.enums import ClassIDType
from logging import getLogger

logger = getLogger(__name__)


def main_spineextract(args):
    outdir = args.outdir
    os.makedirs(outdir, exist_ok=True)
    with open(args.infile, "rb") as f:
        env = load_assetbundle(f)
        objects = [pobj.read() for pobj in env.objects]
        binaries = {
            obj.name: obj
            for obj in objects
            if getattr(obj, "type", None) in {ClassIDType.TextAsset}
        }
        textures = {
            obj.name: obj
            for obj in objects
            if getattr(obj, "type", None) in {ClassIDType.Texture2D}
        }
        spines = set()
        for name in binaries:
            if name.endswith(".atlas") or name.endswith(".skel"):
                spines.add(".".join(name.split(".")[:-1]))
        for spine in spines:
            logger.info("Extracting %s" % spine)
            atlas = binaries.get(spine + ".atlas", None)
            skel = binaries.get(spine + ".skel", None)
            os.makedirs(os.path.join(outdir, spine), exist_ok=True)
            if atlas:
                logger.info("...has Atlas %s" % spine)
                with open(os.path.join(outdir, spine, spine + ".atlas.txt"), "wb") as f:
                    f.write(atlas.script)
                texfiles = [line.strip() for line in atlas.text.split("\n")]
                texfiles = [
                    ".".join(line.split(".")[:-1])
                    for line in texfiles
                    if (line.endswith(".png"))
                ]
                for tex in texfiles:
                    logger.info("...has Texture %s" % tex)
                    textureobj = textures.get(tex, None)
                    if textureobj:
                        textureobj.image.save(os.path.join(outdir, spine, tex + ".png"))
                    else:
                        logger.warning("No texture found for %s" % tex)
            else:
                logger.warning(
                    "No atlas found for %s. Consequnetially, no textures cannot be exported either."
                    % spine
                )

            if skel:
                logger.info("...has Skeleton %s" % spine)
                with open(
                    os.path.join(outdir, spine, spine + ".skel.bytes"), "wb"
                ) as f:
                    f.write(skel.script)
            else:
                logger.warning("No skel found for %s" % spine)



import os


def main_abdecrypt(args):
    from sssekai.crypto.AssetBundle import decrypt

    tree = os.walk(args.indir)
    for root, dirs, files in tree:
        for fname in files:
            file = os.path.join(root, fname)
            if os.path.isfile(file):
                decrypt(open(file, "rb"), open(os.path.join(args.outdir, fname), "wb"))



from typing import BinaryIO
from sssekai.crypto.AssetBundle import has_magic, decrypt

from . import UnityPy, sssekai_get_unity_version
from UnityPy import Environment


def load_assetbundle(file: BinaryIO) -> Environment:
    UnityPy.config.FALLBACK_VERSION_WARNED = True
    UnityPy.config.FALLBACK_UNITY_VERSION = sssekai_get_unity_version()
    stream = file
    if has_magic(file):
        stream = decrypt(file)
    return UnityPy.load(stream)



from Crypto.Cipher import AES


def PKCS7_pad(data: bytes, bs) -> bytes:
    size = bs - len(data) % bs
    return data + bytes([size] * size)


def PKCS7_unpad(data: bytes, bs) -> bytes:
    pad = data[-1]
    return data[:-pad]


def decrypt_aes_cbc(data: bytes, key, iv, unpad=PKCS7_unpad) -> bytes:
    cipher = AES.new(key, AES.MODE_CBC, iv)
    return unpad(cipher.decrypt(data), cipher.block_size)


def encrypt_aes_cbc(data: bytes, key, iv, pad=PKCS7_pad) -> bytes:
    cipher = AES.new(key, AES.MODE_CBC, iv)
    return cipher.encrypt(pad(data, cipher.block_size))


SEKAI_APIMANAGER_KEY = b"g2fcC0ZczN9MTJ61"
SEKAI_APIMANAGER_IV = b"msx3IV0i9XE5uYZ1"


def encrypt(data: bytes) -> bytes:
    return encrypt_aes_cbc(data, SEKAI_APIMANAGER_KEY, SEKAI_APIMANAGER_IV, PKCS7_pad)


def decrypt(data: bytes) -> bytes:
    return decrypt_aes_cbc(data, SEKAI_APIMANAGER_KEY, SEKAI_APIMANAGER_IV, PKCS7_unpad)



from pickle import load, dump
from typing import BinaryIO, List, Mapping, Union
from logging import getLogger
from dataclasses import dataclass, fields, is_dataclass


def fromdict(klass: type, d: Union[Mapping, List]):
    """
    Recursively convert a dictionary to a dataclass instance, if applicable.
    https://stackoverflow.com/a/54769644
    https://gist.github.com/gatopeich/1efd3e1e4269e1e98fae9983bb914f22
    """
    if is_dataclass(klass):
        fieldtypes = {f.name: f.type for f in fields(klass)}
        return klass(**{f: fromdict(fieldtypes[f], d[f]) for f in d})
    if isinstance(d, list) and hasattr(klass, "__args__"):
        return [fromdict(klass.__args__[0], di) for di in d]
    if isinstance(d, dict) and hasattr(klass, "__args__"):
        return {k: fromdict(klass.__args__[1], v) for k, v in d.items()}
    return d


from requests import Session, Response
from msgpack import unpackb, packb

from sssekai import __version__
from sssekai.unity import sssekai_get_unity_version
from sssekai.crypto.APIManager import decrypt, encrypt

logger = getLogger("sssekai.abcache")


@dataclass
class AbCacheConfig:
    app_version: str
    app_platform: str
    app_hash: str


@dataclass
class AbCacheEntry(dict):
    bundleName: str
    cacheFileName: str
    cacheDirectoryName: str
    hash: str
    category: str
    crc: int
    fileSize: int
    dependencies: List[str]
    paths: List[str]
    isBuiltin: bool


@dataclass
class AbCacheIndex(dict):
    version: str
    os: str
    bundles: Mapping[str, AbCacheEntry] = None


@dataclass
class SekaiAppVersion:
    systemProfile: str
    appVersion: str
    multiPlayVersion: str
    appVersionStatus: str
    assetVersion: str
    # These are sadly removed from 3.5.0
    _dataVersion: str = None
    _appHash: str = None
    _assetHash: str = None


@dataclass
class SekaiSystemData:
    serverDate: int
    timezone: str
    profile: str
    maintenanceStatus: str
    appVersions: List[SekaiAppVersion]

    _appVersionDict: dict = None

    @property
    def appVersionDict(self):
        if self._appVersionDict is None:
            self._appVersionDict = {av.appVersion: av for av in self.appVersions}
        return self._appVersionDict


@dataclass
class SekaiGameVersionData:
    profile: str
    assetbundleHostHash: str
    domain: str


@dataclass
class SekaiUserRegistrationData:
    userId: int
    signature: str
    platform: str
    deviceModel: str
    operatingSystem: str
    registeredAt: int


@dataclass
class SekaiUserData:
    userRegistration: SekaiUserRegistrationData
    credential: str
    updatedResources: dict


@dataclass
class SekaiUserAuthData:
    sessionToken: str
    appVersion: str
    multiPlayVersion: str
    dataVersion: str
    assetVersion: str
    removeAssetVersion: str
    assetHash: str
    appVersionStatus: str
    isStreamingVirtualLiveForceOpenUser: bool
    deviceId: str
    updatedResources: dict
    suiteMasterSplitPath: list


@dataclass
class SSSekaiDatabase:
    config: AbCacheConfig = None
    sekai_user_data: SekaiUserData = None
    sekai_user_auth_data: SekaiUserAuthData = None
    sekai_abcache_index: AbCacheIndex = None
    sekai_system_data: SekaiSystemData = None
    sekai_gameversion_data: SekaiGameVersionData = None


class AbCacheBundleNotFoundError(Exception):
    bundleName: str

    def __init__(self, bundleName: str):
        self.bundleName = bundleName
        super().__init__("Bundle not found: %s" % bundleName)


class AbCache(Session):
    database: SSSekaiDatabase

    @property
    def config(self):
        return self.database.config

    @config.setter
    def config(self, v):
        self.database.config = v

    @property
    def SEKAI_APP_VERSION(self):
        return self.config.app_version

    @property
    def SEKAI_APP_PLATFORM(self):
        return self.config.app_platform

    @property
    def SEKAI_APP_HASH(self):
        return self.config.app_hash

    @property
    def SEKAI_ASSET_VERSION(self):
        return self.database.sekai_system_data.appVersionDict[
            self.SEKAI_APP_VERSION
        ].assetVersion

    @property
    def SEKAI_AB_HASH(self):
        return self.database.sekai_user_auth_data.assetHash

    @property
    def SEKAI_AB_HOST_HASH(self):
        return self.database.sekai_gameversion_data.assetbundleHostHash

    @property
    def SEKAI_API_ENDPOINT(self):
        return "https://production-game-api.sekai.colorfulpalette.org"

    @property
    def SEKAI_API_SYSTEM_DATA(self):
        return self.SEKAI_API_ENDPOINT + "/api/system"

    @property
    def SEKAI_API_USER(self):
        return self.SEKAI_API_ENDPOINT + "/api/user"

    @property
    def SEKAI_API_USER_AUTH(self):
        return f"{self.SEKAI_API_USER}/{self.database.sekai_user_data.userRegistration.userId}/auth?refreshUpdatedResources=False"

    @property
    def SEKAI_API_USER_SUITE(self):
        return f"{self.SEKAI_API_ENDPOINT}/api/suite/user/{self.database.sekai_user_data.userRegistration.userId}"

    @property
    def SEKAI_API_MASTER_SUITE(self):
        return f"{self.SEKAI_API_ENDPOINT}/api/suite/master"

    @property
    def SEKAI_API_INFORMATION(self):
        return self.SEKAI_API_ENDPOINT + "/api/information"

    @property
    def SEKAI_API_GAMEVERSION_ENDPOINT(self):
        return "https://game-version.sekai.colorfulpalette.org"

    @property
    def SEKAI_AB_INFO_ENDPOINT(self):
        return f"https://production-{self.SEKAI_AB_HOST_HASH}-assetbundle-info.sekai.colorfulpalette.org/"

    @property
    def SEKAI_AB_ENDPOINT(self):
        return f"https://production-{self.SEKAI_AB_HOST_HASH}-assetbundle.sekai.colorfulpalette.org/"

    @property
    def SEKAI_AB_BASE_PATH(self):
        return f"{self.SEKAI_ASSET_VERSION}/{self.SEKAI_AB_HASH}/{self.config.app_platform}/"

    @property
    def SEKAI_AB_INDEX_PATH(self):
        return f"api/version/{self.SEKAI_ASSET_VERSION}/os/{self.config.app_platform}"

    @property
    def SEKAI_ISSUE_ENDPOINT(self):
        return "https://issue.sekai.colorfulpalette.org"

    @property
    def SEKAI_ISSUE_SIGNATURE_ENDPOINT(self):
        return self.SEKAI_ISSUE_ENDPOINT + "/api/signature"

    @property
    def abcache_index(self):
        return self.database.sekai_abcache_index

    def request_packed(self, method: str, url: str, data: dict = None, **kwargs):
        """Send a request with packed data. Data will be packed and encrypted before sending.

        Args:
            method (str): HTTP method
            url (str): URL
            data (dict, optional): Payload data. Defaults to None.

        Returns:
            Response: Response object
        """
        if data is not None:
            data = packb(data)
            data = encrypt(data)
        resp = self.request(method=method, url=url, data=data, **kwargs)
        resp.raise_for_status()
        return resp

    @staticmethod
    def response_to_dict(resp: Response):
        """Decrypt and unpack a response content to a dictionary.

        Args:
            resp (Response): Response object

        Returns:
            dict: Decrypted and unpacked data
        """
        data = decrypt(resp.content)
        data = unpackb(data)
        return data

    def _update_signatures(self):
        logger.debug("Updating signatures")
        resp = self.request_packed("POST", self.SEKAI_ISSUE_SIGNATURE_ENDPOINT)
        self.headers["Cookie"] = resp.headers["Set-Cookie"]
        # HACK: Per RFC6265, Cookies should not be visible to subdomains since it's not set with Domain attribute (https://github.com/psf/requests/issues/2576)
        # But the other endpoints uses it nontheless. So we have to set it manually.

    def _update_user_data(self):
        logger.debug("Updating user data")
        payload = {
            "platform": self.headers["X-Platform"],
            "deviceModel": self.headers["X-DeviceModel"],
            "operatingSystem": self.headers["X-OperatingSystem"],
        }
        resp = self.request_packed("POST", self.SEKAI_API_USER, data=payload)
        data = self.response_to_dict(resp)
        self.database.sekai_user_data = fromdict(SekaiUserData, data)

    def _update_user_auth_data(self):
        logger.debug("Updating user auth data")
        payload = {
            "credential": self.database.sekai_user_data.credential,
            "deviceId": None,
        }
        resp = self.request_packed("PUT", self.SEKAI_API_USER_AUTH, data=payload)
        data = self.response_to_dict(resp)
        self.database.sekai_user_auth_data = fromdict(SekaiUserAuthData, data)

    def _update_system_data(self):
        logger.debug("Updating system data")
        resp = self.request_packed("GET", self.SEKAI_API_SYSTEM_DATA)
        data = self.response_to_dict(resp)
        self.database.sekai_system_data = fromdict(SekaiSystemData, data)

    def _update_gameversion_data(self):
        logger.debug("Updating game version data")
        resp = self.request_packed(
            "GET",
            self.SEKAI_API_GAMEVERSION_ENDPOINT
            + "/"
            + self.SEKAI_APP_VERSION
            + "/"
            + self.SEKAI_APP_HASH,
        )
        data = self.response_to_dict(resp)
        self.database.sekai_gameversion_data = fromdict(SekaiGameVersionData, data)

    def _update_abcache_index(self) -> AbCacheIndex:
        logger.debug("Updating Assetbundle index")
        resp = self.request_packed(
            "GET", self.SEKAI_AB_INFO_ENDPOINT + self.SEKAI_AB_INDEX_PATH
        )
        data = self.response_to_dict(resp)
        self.database.sekai_abcache_index = fromdict(AbCacheIndex, data)
        return self.database.sekai_abcache_index

    def __init__(self, config: AbCacheConfig) -> None:
        super().__init__()
        self.database = SSSekaiDatabase()
        self.config = config
        self.config.app_platform = self.config.app_platform.lower()
        self.headers.update(
            {
                "Accept": "application/octet-stream",
                "Content-Type": "application/octet-stream",
                "Accept-Encoding": "deflate, gzip",
                "User-Agent": "UnityPlayer/%s" % sssekai_get_unity_version(),
                "X-Platform": self.config.app_platform.capitalize(),
                "X-DeviceModel": "sssekai/%s" % __version__,
                "X-OperatingSystem": self.config.app_platform.capitalize(),
                "X-Unity-Version": sssekai_get_unity_version(),
                "X-App-Version": self.SEKAI_APP_VERSION,
                "X-App-Hash": self.SEKAI_APP_HASH,
            }
        )

    def update_download_headers(self):
        """Update headers for downloading assetbundles *ONLY*. Functionalities related to user-level data (e.g. Master Data, etc) won't
        be available.

        Returns:
            dict: Updated headers
        """
        self._update_signatures()
        return self.headers

    def update_client_headers(self):
        """Update headers for client-level functionalities. This includes user-level data (e.g. Master Data, etc) and assetbundle data.

        Returns:
            _type_: _description_
        """
        logger.debug("Updating metadata")
        logger.debug(
            "Set App version: %s (%s), hash=%s"
            % (self.config.app_version, self.config.app_platform, self.SEKAI_APP_HASH)
        )
        self._update_signatures()
        self._update_system_data()
        version_newest = self.database.sekai_system_data.appVersions[-1]
        logger.debug("Newest App version: %s" % version_newest)
        if version_newest.appVersion != self.SEKAI_APP_VERSION:
            logger.warning("App version mismatch. This may cause issues.")
        self._update_gameversion_data()
        self._update_user_data()
        self._update_user_auth_data()
        self.headers.update(
            {
                "X-Data-Version": self.database.sekai_user_auth_data.dataVersion,
                "X-Asset-Version": self.database.sekai_user_auth_data.assetVersion,
                "X-Session-Token": self.database.sekai_user_auth_data.sessionToken,
            }
        )
        return self.headers

    def update(self):
        self.update_client_headers()
        self._update_abcache_index()
        logger.debug("Sekai AssetBundle version: %s" % self.SEKAI_ASSET_VERSION)
        logger.debug("Sekai AssetBundle host hash: %s" % self.SEKAI_AB_HOST_HASH)

    def save(self, f: BinaryIO):
        logger.debug("Saving cache")
        dump(self.database, f)

    def load(self, f: BinaryIO):
        logger.debug("Loading cache")
        self.database = load(f)

    def __repr__(self) -> str:
        return (
            f"<AbCache config={self.config} bundles={len(self.abcache_index.bundles)}>"
        )

    def get_entry_by_bundle_name(self, bundleName: str) -> AbCacheEntry:
        return self.abcache_index.bundles.get(bundleName, None)

    def get_entry_download_url(self, entry: AbCacheEntry):
        return self.SEKAI_AB_ENDPOINT + self.SEKAI_AB_BASE_PATH + entry.bundleName

    def get_or_update_dependency_tree_flatten(self, bundleName: str, deps: set = None):
        """Get a flattened set of asset dependency bundle names (including itself) for a given entry.

        Returned order is topologically sorted.

        if 'deps' is provided, it will be used as the initial set of dependencies.
        """
        deps = deps or set()

        def dfs(bundleName):
            for dep in self.abcache_index.bundles[bundleName].dependencies:
                if dep not in deps:
                    dfs(dep)
                    deps.add(dep)

        dfs(bundleName)
        deps.add(bundleName)
        return deps



from io import BytesIO
from sssekai.unity.AssetBundle import load_assetbundle
from sssekai.fmt.moc3 import read_moc3
import sys, os
from UnityPy.enums import ClassIDType


def main_moc3paths(args):
    ParameterNames = set()
    PartNames = set()
    tree = os.walk(args.indir)
    for root, dirs, files in tree:
        for fname in files:
            file = os.path.join(root, fname)
            with open(file, "rb") as f:
                env = load_assetbundle(f)
                for obj in env.objects:
                    if obj.type == ClassIDType.TextAsset:
                        data = obj.read()
                        out_name: str = data.name
                        if out_name.endswith(".moc3"):
                            parts, parameters = read_moc3(
                                BytesIO(data.script.tobytes())
                            )
                            ParameterNames.update(parameters)
                            PartNames.update(parts)
    from zlib import crc32

    print("NAMES_CRC_TBL = {")
    for name in sorted(list(PartNames)):
        fullpath = "Parts/" + name
        print('    %d:"%s",' % (crc32(fullpath.encode("utf-8")), fullpath))
    for name in sorted(list(ParameterNames)):
        fullpath = "Parameters/" + name
        print('    %d:"%s",' % (crc32(fullpath.encode("utf-8")), fullpath))
    print("}")



import zipfile
import UnityPy
import argparse
import re

from io import BytesIO

HASHREGEX = re.compile(b"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}")


def dump_axml_stringpool(f: BytesIO):
    read_int = lambda nbytes: int.from_bytes(f.read(nbytes), "little")
    f.seek(8)
    hdr_type, hdr_size, size = read_int(2), read_int(2), read_int(4)
    n_strings = read_int(4)
    unk = read_int(4)
    flags = read_int(4)
    is_utf8 = flags & (1 << 8)
    string_offset = read_int(4) + 8
    unk = read_int(4)
    offsets = [read_int(4) for _ in range(n_strings)]
    for offset in offsets:
        f.seek(string_offset + offset)
        string_len = read_int(2)
        if is_utf8:
            string = f.read(string_len).decode("utf-8")
        else:
            string = f.read(string_len * 2).decode("utf-16-le")
        yield string


def enum_candidates(zip_file, filter):
    return (
        (f, zip_file.open(f), zip_file) for f in zip_file.filelist if filter(f.filename)
    )


def enum_package(zip_file):
    yield zip_file
    for f in zip_file.filelist:
        if f.filename.lower().endswith(".apk"):
            yield zipfile.ZipFile(zip_file.open(f))


def main_apphash(args):
    env = UnityPy.Environment()
    if not args.apk_src or args.fetch:
        from requests import get

        src = BytesIO()
        print("Fetching latest game package from APKPure")
        resp = get(
            "https://d.apkpure.net/b/XAPK/com.sega.pjsekai?version=latest", stream=True
        )
        size = resp.headers.get("Content-Length", -1)
        for chunck in resp.iter_content(chunk_size=2**10):
            src.write(chunck)
            print("Downloading %d/%s" % (src.tell(), size), end="\r")
        print()
        src.seek(0)
    else:
        src = open(args.apk_src, "rb")
    with zipfile.ZipFile(src, "r") as zip_ref:
        manifests = [
            manifest
            for package in enum_package(zip_ref)
            for manifest in enum_candidates(
                package, lambda fn: fn == "AndroidManifest.xml"
            )
        ]
        manifest = manifests[0][1]
        manifest_strings = list(dump_axml_stringpool(manifest))
        for i in range(len(manifest_strings)):
            if "STAMP_TYPE_" in manifest_strings[i]:
                print("Package Type:", manifest_strings[i])
                print("Version:", manifest_strings[i - 2])
                break
        candidates = [
            candidate
            for package in enum_package(zip_ref)
            for candidate in enum_candidates(
                package,
                lambda fn: fn.split("/")[-1]
                in {
                    "6350e2ec327334c8a9b7f494f344a761",  # PJSK Android
                    "c726e51b6fe37463685916a1687158dd",  # PJSK iOS
                },
            )
        ]
        for candidate, stream, _ in candidates:
            env.load_file(stream)
        for obj in env.objects:
            obj = obj.read()
            hashStr = HASHREGEX.finditer(obj.raw_data)
            for m in hashStr:
                print(m.group().decode(), obj.name)



import os


def main_apidecrypt(args):
    from sssekai.crypto.APIManager import decrypt
    import json, msgpack

    data = open(args.infile, "rb").read()
    with open(args.outfile, "w", encoding="utf-8") as fout:
        plain = decrypt(data)
        msg = msgpack.unpackb(plain)
        json.dump(msg, fout, indent=4, ensure_ascii=False)



from base64 import b64decode, b64encode
from struct import unpack as s_unpack
from io import BytesIO
from collections import defaultdict
import math, gzip
import msgpack

read_int = lambda stream, nbytes, signed=False: int.from_bytes(
    stream.read(nbytes), "little", signed=signed
)
read_float = lambda stream: s_unpack("<f", stream.read(4))[0]


class SSEDataLengthOutOfRangeException(Exception):
    needed: int
    current: int

    def __init__(self, needed, current):
        self.needed = needed
        self.current = current
        super().__init__(f"Needed {needed} bytes, but only {current} bytes available.")


# Sekai_Streaming_StreamingCommon__CheckHeader
def decode_buffer_base64(buffer: bytes) -> tuple[int, bytes]:
    """Decodes the 'RTVL' SSE Message payload into a header signature and data.

    Args:
        buffer (bytes): Encoded buffer

    Raises:
        SSEDataLengthOutOfRangeException: If the buffer length does not match the expected length.

    Returns:
        (int, bytes): Header Signature and decoded data payload
    """
    # They really want you to believe it's Base64...
    # [4 bytes: RTVL][6 bytes : length in hex][Base64 T/F][Split T/F][3 bytes: Signature][data, Base64]
    # ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^ 15 bytes
    # Sekai_Streaming_StreamingCommon__cctor
    SYM_HEADER = b"RTVL"
    SYM_TRUE = b"T"
    SYM_FALSE = b"F"
    stream = BytesIO(buffer)
    assert stream.read(4) == SYM_HEADER, "bad header"
    encoded_length = stream.read(6)
    encoded_length = int(encoded_length, 16)
    is_base64_encoded = stream.read(1) == SYM_TRUE
    is_split = stream.read(1) == SYM_TRUE
    header_signature = int(stream.read(3).decode(), 10)
    data = stream.read()
    current_length = len(data) + 15
    if current_length != encoded_length:
        raise SSEDataLengthOutOfRangeException(encoded_length, current_length)
    if is_base64_encoded:
        data = b64decode(data)
    return header_signature, data


# Sekai_Streaming_SubscribeDecoder__Deserialize
def decode_buffer_payload(buffer: bytes) -> tuple[int, bytes]:
    """Decodes the decoded data payload (in binary form) into a decoder signature and payload data.

    Args:
        buffer (bytes): Decoded buffer

    Returns:
        (int, bytes): Decoder signature and decoded payload data. The signature should match the Header Signature.
    """
    stream = BytesIO(buffer)
    decoder_signature = read_int(stream, 1)
    unk1 = read_int(stream, 4)
    if not unk1:
        unk2 = read_int(stream, 4)
        if unk2:
            payload = stream.read()
            payload = gzip.decompress(payload)
            return decoder_signature, payload
    return decoder_signature, stream.read()


# Sekai_Streaming_StreamingData__Deserialize
def decode_streaming_data(
    version: tuple, decoder_signature, buffer, strict=True
) -> dict:
    """Decodes the streaming data payload into a dictionary.

    Args:
        version (tuple): RLA version
        decoder_signature (int): Decoder signature
        buffer (bytes): Decoded payload data
        strict (bool, optional): If False, incomplete packets will be returned as is. Defaults to True.

    Raises:
        Exception: If the packet is incomplete and strict is True.

    Returns:
        dict: Parsed streaming data payload
    """
    stream = BytesIO(buffer)
    n_mask_offset = read_int(stream, 4)
    n_init_pos = stream.tell()
    stream.seek(n_mask_offset)
    n_mask_length = read_int(stream, 2)
    bitmask = stream.read(n_mask_length)
    # assert not stream.read() # EOF
    stream.seek(n_init_pos)
    # CP_Serialize_SerializableValueSet
    mask_i = -1

    def get_next_mask():
        nonlocal mask_i
        mask_i += 1
        return bitmask[mask_i // 8] & (1 << (mask_i % 8)) != 0

    get_next_pred = lambda: get_next_mask() | (get_next_mask() << 1)
    get_next_byte = lambda: [
        lambda: 0,
        lambda: 1,
        lambda: -1,
        lambda: read_int(stream, 1),
    ][
        get_next_pred()
    ]()  # ReadByte
    get_next_ushort = lambda: [
        lambda: 0.0,
        lambda: 0.0,
        lambda: read_int(stream, 1),
        lambda: read_int(stream, 2),
    ][
        get_next_pred()
    ]()  # ReadUShort
    get_next_short = lambda: [
        lambda: 0.0,
        lambda: 0.0,
        lambda: read_int(stream, 1, True),
        lambda: read_int(stream, 2),
    ][
        get_next_pred()
    ]()  # ReadShort
    get_next_int = lambda: [
        lambda: 0,
        lambda: read_int(stream, 1),
        lambda: read_int(stream, 1, True),
        lambda: read_int(stream, 4),
    ][
        get_next_pred()
    ]()  # ReadInt
    get_next_long = lambda: [
        lambda: 0,
        lambda: read_int(stream, 1),
        lambda: read_int(stream, 2),
        lambda: read_int(stream, 8),
    ][
        get_next_pred()
    ]()  # ReadLong
    get_next_float = lambda: [
        lambda: 0.0,
        lambda: 0.0,
        lambda: 3.4028e38,
        lambda: read_float(stream),
    ][
        get_next_pred()
    ]()  # ReadSingle
    get_next_vector3 = lambda: (
        get_next_float(),
        get_next_float(),
        get_next_float(),
    )  # ReadVector3
    get_next_ushort_vector3 = lambda: (
        get_next_ushort() * 0.01,
        get_next_ushort() * 0.01,
        get_next_ushort() * 0.01,
    )  # ReadUShortVector3
    get_next_tiny_int = lambda type: {
        "+": lambda: read_int(stream, 2),
        "*": lambda: read_int(stream, 1, True),
        ")": lambda: read_int(stream, 1),
    }[
        type
    ]()  # ReadTinyInt
    get_next_string = lambda: (
        stream.read(get_next_tiny_int(chr(read_int(stream, 1)))).decode()
        if get_next_pred()
        else None
    )  # ReadString

    # HACK: Workaround for incomplete packets.
    def gen_exception_handler(generator):
        try:
            for item in generator:
                yield item
        except Exception as e:
            # XXX: This shouldn't happen unless the packet is corrupted.
            if strict:
                raise e
            else:
                return

    gen_get_next_array = lambda reader: (
        reader() for _ in range(get_next_int())
    )  # ReadArray<T>
    get_next_array = lambda reader: list(
        gen_exception_handler(gen_get_next_array(reader))
    )
    # Sekai_Streaming_StreamingData__Deserialize
    assert get_next_byte() == decoder_signature, "bad signature"
    compress_type = get_next_int()
    sequence_no = get_next_int()
    target_time = get_next_long()
    # StreamingData$$ReadValue implementations
    deg_to_rad = lambda value: tuple(deg * (math.pi / 180) for deg in value)
    get_next_deg_as_rad = lambda: deg_to_rad(
        get_next_ushort_vector3()
    )  # Eulers [0, 2pi]
    get_next_pose_data = lambda: {
        "bodyPosition": get_next_vector3(),
        "bodyRotation": get_next_deg_as_rad(),  # Eulers
        **(
            {
                "musicItemPropPosition": get_next_vector3(),
                "musicItemPropRotation": get_next_deg_as_rad(),
            }
            if version >= (1, 1)
            else {}
        ),
        "boneDatas": get_next_array(get_next_deg_as_rad),  # Eulers [0, 2pi]
        "shapeDatas": get_next_array(get_next_float),  # [0,100]
        **(
            {"propBoneDatas": get_next_array(get_next_deg_as_rad)}  # Eulers [0, 2pi]
            if version >= (1, 1)
            else {}
        ),
        "heightOffset": get_next_short() * 0.01,
        "isActive": get_next_mask(),
        "useActiveFx": get_next_mask(),
        **({"isEyeLookAt": get_next_mask()} if version >= (1, 4) else {}),
    }
    match decoder_signature:
        case 0:
            # Sekai_Streaming_MotionData
            timeStamps = get_next_array(get_next_long) if get_next_pred() == 2 else None
            poses = get_next_array(get_next_pose_data)
            return {"type": "MotionData", "timeStamps": timeStamps, "poses": poses}
        case 1:
            # Sekai_Streaming_MotionCaptureData
            read_character_capture_data = lambda: {
                "id": get_next_int(),
                "timestamp": get_next_long(),
                "pose": get_next_pose_data(),
            }
            data = get_next_array(read_character_capture_data)
            return {"type": "MotionCaptureData", "data": data}
        case 2:
            # Sekai_Streaming_SoundData
            channels = get_next_int()
            sample_rate = get_next_int()
            data_length = get_next_int()
            if compress_type != 1:
                data = stream.read(data_length)  # Raw sampling data
                return {
                    "type": "SoundData",
                    "channels": channels,
                    "sampleRate": sample_rate,
                    "encoding": "raw",
                    "data": b64encode(data).decode(),
                }
            else:
                data = stream.read(data_length)  # HCA w/o metadata
                return {
                    "type": "SoundData",
                    "channels": channels,
                    "sampleRate": sample_rate,
                    "encoding": "hca",
                    "data": b64encode(data).decode(),
                }
        case 3:
            # Sekai_Streaming_StatusData
            read_stage_status = lambda: {
                "liveState": get_next_byte(),
                "lightIntensity": get_next_float(),
                "gayaVolume": get_next_float(),
                "cheerVolume": get_next_float(),
                "characterSpotlightIndex": get_next_int(),
                "characterSpotlightIntensity": get_next_float(),
                "stageSetlistIndex": get_next_int(),
                "musicSetlistIndex": get_next_int(),
                "musicTargetTime": get_next_long(),
                "musicStartTime": get_next_float(),
                "seId": get_next_int(),
                "seStartTime": get_next_long(),
                "timeStamp": get_next_long(),
                "unk0": get_next_byte(),
                "unk1": get_next_byte(),
                "playTimelineId": get_next_int(),
                **(
                    {
                        "screenFadeColorR": get_next_byte(),
                        "screenFadeColorG": get_next_byte(),
                        "screenFadeColorB": get_next_byte(),
                        "screenFade": get_next_float(),
                        "characterFormationRotate": get_next_int(),
                        "stageCenterPosition": get_next_vector3(),
                        "playerAvatarStartPosition": get_next_vector3(),
                    }
                    if version >= (1, 4)
                    else {}
                ),
            }
            read_character_status = lambda: {
                "costumeIndex": get_next_int(),
                **({"useFx": get_next_mask()} if version >= (1, 2) else {}),
                "timeStamp": get_next_long(),
            }
            stage_status_list = get_next_array(read_stage_status)
            stage_status_length = get_next_int()
            charcter_status_list = [
                get_next_array(read_character_status)
                for _ in range(stage_status_length)
            ]
            return {
                "type": "StatusData",
                "stageStatus": stage_status_list,
                "characterStatus": charcter_status_list,
            }
        case 4:
            # Sekai_Streaming_VirtualLiveMessageData
            message_id = get_next_int()
            user_id = get_next_string()
            data_length = get_next_int()
            data = stream.read(data_length)
            # CP_BinarySerializer__Deserialize_OtherRoomMessageData
            # CP_BinarySerializer__Deserialize_OtherRoomActionData
            # NOTE: The msgpack payload does not contain the message type
            data = msgpack.unpackb(data)
            return {
                "type": "VirtualLiveMessageData",
                "messageId": message_id,
                "userId": user_id,
                "data": data,
            }
        case 5:
            # Sekai_Streaming_ComplementInfoData
            info_type = get_next_int()
            info_data = get_next_string()
            return {
                "type": "ComplementInfoData",
                "infoType": info_type,
                "infoData": info_data,
            }
    return {"type": "Unknown", "data": buffer}


result = defaultdict(dict)


def read_rla(src: BytesIO, version=(1, 0), strict=True) -> dict:
    """Parses the Sekai RLA file format used in 'streaming_live/archive' assets.

    Args:
        src (BytesIO): Source RLA file stream
        version (tuple, optional): RLA version, found in respective RLH (JSON) header files. i.e. one of (1,0), (1,1), (1,2), (1,3), (1,4) Defaults to (1,0).
        strict (bool, optional): If False, incomplete packets will be returned as is. Defaults to True.

    Returns:
        dict: Parsed RLA data. The dictionary is sorted by the frame ticks.
    """
    result = defaultdict(dict)

    def read_frames():
        ticks = read_int(src, 8)
        if ticks:
            buffer_length = read_int(src, 4)
            buffer = src.read(buffer_length)
            header_signature, data = decode_buffer_base64(buffer)
            decoder_signature, data = decode_buffer_payload(data)
            assert (
                header_signature == decoder_signature
            ), "mismatching signature (header/decoder). packet may be corrupt"
            payload = decode_streaming_data(version, decoder_signature, data, strict)
            result[ticks].setdefault(payload["type"], list()).append(payload)
            return True
        return False

    while read_frames():
        pass
    result = dict(sorted(result.items(), key=lambda x: x[0]))
    return result


if __name__ == "__main__":
    from timeit import timeit

    fp = open(
        r"C:\Users\mos9527\Desktop\sekai_streaming\sekai_30_00000060.rla.bytes", "rb"
    )
    buffer = fp.read()
    fp = BytesIO(buffer)
    print("took %.2f seconds" % timeit("read_rla(fp, (1,0))", globals=globals()))



from io import BytesIO
from typing import BinaryIO

SEKAI_AB_MAGIC = b"\x10\x00\x00\x00"


def decrypt_headaer_inplace(header: bytearray):
    assert len(header) == 128
    for i in range(0, 128, 8):
        for j in range(5):
            header[i + j] = ~header[i + j] & 0xFF
    return header


def decrypt(fin: BinaryIO, fout: BinaryIO = None) -> None:
    if not fout:
        fout = BytesIO()
    magic = fin.read(4)
    if magic == SEKAI_AB_MAGIC:
        fout.write(decrypt_headaer_inplace(bytearray(fin.read(128))))
        while block := fin.read(65536):
            fout.write(block)
    else:
        fin.seek(0)
        while block := fin.read(65536):
            fout.write(block)
    return fout


def encrypt(fin, fout):
    raise NotImplementedError


def has_magic(fin: BinaryIO) -> bool:
    magic = fin.read(4)
    fin.seek(0)
    return magic == SEKAI_AB_MAGIC



from sssekai.unity.AnimationClip import read_animation
from UnityPy.classes import AnimationClip
from logging import getLogger

logger = getLogger(__name__)


# Thanks! https://github.com/Perfare/UnityLive2DExtractor/blob/master/UnityLive2DExtractor/CubismMotion3Converter.cs
def unity_animation_clip_to_motion3(
    animationClip: AnimationClip, pathTable: dict
) -> dict:
    """Convert Unity AnimationClip to Live2D Motion3 format

    Args:
        animationClip (AnimationClip): animationClip
        pathTable (dict): CRC32 to Live2D path table

    Returns:
        dict: Live2D Motion3 data
    """
    motion = {
        "Version": 3,
        "Meta": {
            "Name": animationClip.m_Name,
            "Duration": animationClip.m_MuscleClip.m_StopTime,
            "Fps": animationClip.m_SampleRate,
            "Loop": True,
            "AreBeziersRestricted": True,
            "CurveCount": 0,
            "UserDataCount": 0,
            "TotalPointCount": 0,
            "TotalSegmentCount": 0,
            "TotalUserDataSize": 0,
        },
        "Curves": [],
        "UserData": [],
    }
    animation = read_animation(animationClip)
    floatCurves = animation.FloatTracks
    motion["Meta"]["CurveCount"] = len(floatCurves)
    for track in [track for path in floatCurves.values() for track in path.values()]:
        segments = list()
        segments.append(0)
        segments.append(track.Curve[0].value)
        curveIndex = 1
        while curveIndex < len(track.Curve):
            curve = track.Curve[curveIndex]
            preCurve = track.Curve[curveIndex - 1]
            if abs(curve.time - preCurve.time - 0.01) < 0.0001:
                nextCurve = track.Curve[curveIndex + 1]
                if nextCurve.value == curve.value:
                    segments.append(3)  # InverseSteppedSegment
                    segments.append(nextCurve.time)
                    segments.append(nextCurve.value)
                    motion["Meta"]["TotalPointCount"] += 1
                    motion["Meta"]["TotalSegmentCount"] += 1
                    curveIndex += 1
                    continue
            if curve.inSlope == float("+inf"):
                segments.append(2)  # SteppedSegment
                segments.append(curve.time)
                segments.append(curve.value)
                motion["Meta"]["TotalPointCount"] += 1
            elif preCurve.outSlope == 0 and abs(curve.inSlope) < 0.0001:
                segments.append(0)  # LinearSegment
                segments.append(curve.time)
                segments.append(curve.value)
                motion["Meta"]["TotalPointCount"] += 1
            else:
                tangentLength = (curve.time - preCurve.time) / 3
                segments.append(1)  # BezierSegment
                segments.append(preCurve.time + tangentLength)
                segments.append(preCurve.outSlope * tangentLength + preCurve.value)
                segments.append(curve.time - tangentLength)
                segments.append(curve.value - curve.inSlope * tangentLength)
                segments.append(curve.time)
                segments.append(curve.value)
                motion["Meta"]["TotalPointCount"] += 3
            motion["Meta"]["TotalSegmentCount"] += 1
            curveIndex += 1
        path = track.Path
        if path in pathTable:
            target, id = pathTable[path].split("/")
            if target == "Parameters":
                target = "Parameter"
            if target == "Parts":
                target = "PartOpacity"
        else:
            logger.warning("Failed to bind path CRC %s to any Live2D path" % path)
            target, id = "PartOpacity", str(path)
        motion["Curves"].append({"Target": target, "Id": id, "Segments": segments})
    for event in animationClip.m_Events:
        motion["UserData"].append({"time": event.time, "value": event.data})
        motion["Meta"]["UserDataCount"] += 1
        motion["Meta"]["TotalUserDataSize"] += len(event.data)
    return motion



from collections import OrderedDict
from typing import Dict, List, Tuple
from UnityPy.enums import ClassIDType
from UnityPy.classes import AnimationClip
from UnityPy.math import Matrix4x4, Quaternion, Vector3
from dataclasses import dataclass
from enum import IntEnum
from logging import getLogger

logger = getLogger(__name__)


# taken from https://github.com/AssetRipper/AssetRipper
class TransformType(IntEnum):
    _None = 0
    Translation = 1
    Rotation = 2
    Scaling = 3
    EulerRotation = 4


def DimensionOfTransformType(type: TransformType):
    if type != TransformType.Rotation:  # Quaternion
        return 3
    return 4  # XYZ


@dataclass
class KeyFrame:
    time: float
    value: float | Vector3 | Quaternion
    inSlope: float | Vector3 | Quaternion = 0
    outSlope: float | Vector3 | Quaternion = 0
    coeff: float = 0


@dataclass
class Track:
    Attribute: int | str
    Path: int | str
    Curve: List[KeyFrame]

    def __init__(self) -> None:
        self.Curve = list()

    def add_keyframe(self, keyframe: KeyFrame):
        self.Curve.append(keyframe)


class Animation:
    FloatTracks: Dict[
        int, Dict[int, Track]
    ]  # Path crc hash - Attribute crc hash - Track
    TransformTracks: Dict[
        TransformType, Dict[int, Track]
    ]  # TransformType - Path crc hash - Track
    Framerate: int
    Duration: int

    def __init__(self) -> None:
        self.Framerate = 60
        self.FloatTracks = dict()
        self.TransformTracks = dict()
        self.TransformTracks[TransformType.EulerRotation] = dict()
        self.TransformTracks[TransformType.Rotation] = dict()
        self.TransformTracks[TransformType.Translation] = dict()
        self.TransformTracks[TransformType.Scaling] = dict()


def read_animation(animationClip: AnimationClip) -> Animation:
    """Reads AnimationClip data and converts it to a list of Tracks

    Args:
        animationClip (AnimationClip): animationClip

    Returns:
        List[Track]: List of Tracks
    """
    m_Clip = animationClip.m_MuscleClip.m_Clip
    streamedFrames = m_Clip.m_StreamedClip.ReadData()
    m_ClipBindingConstant = animationClip.m_ClipBindingConstant
    animationTracks = Animation()
    animationTracks.Framerate = animationClip.m_SampleRate
    animationTracks.Duration = animationClip.m_MuscleClip.m_StopTime

    def get_transform_track(
        path: int, type: TransformType
    ):  # TransformType is attribute. transposed since there are only TRS
        if not path in animationTracks.TransformTracks[type]:
            animationTracks.TransformTracks[type][path] = Track()
        return animationTracks.TransformTracks[type][path]

    def get_float_track(path: int, attribute: int = 0) -> Track:
        if not path in animationTracks.FloatTracks:
            animationTracks.FloatTracks[path] = dict()
        if not attribute in animationTracks.FloatTracks[path]:
            animationTracks.FloatTracks[path][attribute] = Track()
        return animationTracks.FloatTracks[path][attribute]

    def add_float_curve_data(binding, time, value, inSlope, outSlope, coeff):
        track = get_float_track(binding.path, binding.attribute)
        track.Attribute = binding.attribute
        track.Path = binding.path
        track.add_keyframe(KeyFrame(time, value, inSlope, outSlope, coeff))

    def add_transform_curve_data(
        binding, time, datas: List[Tuple[float, float, float]]
    ):  # value, inSlope, outSlope
        type = TransformType(binding.attribute)
        dimension = DimensionOfTransformType(binding.attribute)
        if dimension == 3:
            frame = KeyFrame(
                time,
                Vector3(datas[0][0], datas[1][0], datas[2][0]),
                Vector3(datas[0][1], datas[1][1], datas[2][1]),
                Vector3(datas[0][2], datas[1][2], datas[2][2]),
            )
        else:
            frame = KeyFrame(
                time,
                Quaternion(datas[0][0], datas[1][0], datas[2][0], datas[3][0]),
                Quaternion(datas[0][1], datas[1][1], datas[2][1], datas[3][1]),
                Quaternion(datas[0][2], datas[1][2], datas[2][2], datas[3][2]),
            )
        track = get_transform_track(binding.path, type)
        track.add_keyframe(frame)

    def get_next_curve_index(current, keyList):
        i = current
        while i < len(keyList) and keyList[current].index == keyList[i].index:
            i += 1
        return i

    for frame in streamedFrames:
        curveIndex = 0
        while curveIndex < len(frame.keyList):
            curveKey = frame.keyList[curveIndex]
            binding = m_ClipBindingConstant.FindBinding(curveKey.index)
            if binding.typeID == ClassIDType.Transform:
                transformType = TransformType(binding.attribute)
                dimension = DimensionOfTransformType(transformType)
                curveData = []
                for _ in range(dimension):
                    curveKey = frame.keyList[curveIndex]
                    curveData.append(
                        (
                            curveKey.value,
                            getattr(curveKey, "inSlope", 0),
                            curveKey.outSlope,
                        )
                    )
                    curveIndex = get_next_curve_index(curveIndex, frame.keyList)
                add_transform_curve_data(binding, frame.time, curveData)
            else:
                add_float_curve_data(
                    binding,
                    frame.time,
                    curveKey.value,
                    getattr(curveKey, "inSlope", 0),
                    curveKey.outSlope,
                    curveKey.coeff,
                )
                curveIndex = get_next_curve_index(curveIndex, frame.keyList)

    m_DenseClip = m_Clip.m_DenseClip
    streamCount = m_Clip.m_StreamedClip.curveCount
    for frameIndex in range(0, m_DenseClip.m_FrameCount):
        time = m_DenseClip.m_BeginTime + frameIndex / m_DenseClip.m_SampleRate
        frameOffset = frameIndex * m_DenseClip.m_CurveCount
        curveIndex = 0
        while curveIndex < m_DenseClip.m_CurveCount:
            index = streamCount + curveIndex
            framePosition = frameOffset + curveIndex
            binding = m_ClipBindingConstant.FindBinding(index)
            if binding.typeID == ClassIDType.Transform:
                transformType = TransformType(binding.attribute)
                dimension = DimensionOfTransformType(transformType)
                curveData = []
                for i in range(dimension):
                    curveData.append(
                        (m_DenseClip.m_SampleArray[framePosition + i], 0, 0)
                    )
                add_transform_curve_data(binding, time, curveData)
                curveIndex += dimension
            else:
                add_float_curve_data(
                    binding, time, m_DenseClip.m_SampleArray[framePosition], 0, 0, 0
                )
                curveIndex += 1

    m_ConstantClip = m_Clip.m_ConstantClip
    denseCount = m_Clip.m_DenseClip.m_CurveCount
    time2 = 0
    for _ in range(0, 2):  # first and last frame
        curveIndex = 0
        while curveIndex < len(m_ConstantClip.data):
            index = streamCount + denseCount + curveIndex
            binding = m_ClipBindingConstant.FindBinding(index)
            if binding.typeID == ClassIDType.Transform:
                transformType = TransformType(binding.attribute)
                dimension = DimensionOfTransformType(transformType)
                curveData = []
                for i in range(dimension):
                    curveData.append((m_ConstantClip.data[curveIndex + i], 0, 0))
                add_transform_curve_data(binding, time2, curveData)
                curveIndex += dimension
            else:
                add_float_curve_data(
                    binding, time2, m_ConstantClip.data[curveIndex], 0, 0, 0
                )
                curveIndex += 1
        time2 = animationClip.m_MuscleClip.m_StopTime
    return animationTracks
